from math import sin, cos, radians
from typing import Callable

from ...settings import Settings
from ...util import Ref, computed, LayoutDescriptor, KeyColumnsTuple, KeyGridTuple

def build_layout_descriptor(settings: Settings) -> LayoutDescriptor:
    #region Key layouts

    TOP_ROW = 0
    LOW_ROW = 2
    LOW_COMPOUND_ROW = 1

    # [keys], label, rowSpan?, numberLabel?
    MAIN_ROWS_KEYS_STAGGERED_LEFT: KeyColumnsTuple = (
        (
            (["^-"], "^"), # ∧
            (["^-", "+-"], ""),
            (["+-"], "+"),
        ), (
            (["^-", "S-"], ""),
            (["^-", "+-", "S-"], ""),
            (["+-", "S-"], ""),
        ), (
            (["S-"], "S", 3, "1"),
        ), (
            (["T-"], "T", 1, "2"),
            (["T-", "K-"], ""),
            (["K-"], "K"),
        ), (
            (["P-"], "P", 1, "3"),
            (["P-", "W-"], ""),
            (["W-"], "W"),
        ), (
            (["H-"], "H", 1, "4"),
            (["H-", "R-"], ""),
            (["R-"], "R"),
        ), (
            (["H-", "&-"], ""),
            (["H-", "R-", "&-"], ""),
            (["R-", "&-"], ""),
        # ), (
        #     (["&-"], "&&", 3), # ＋
        # ), (
        #     (["&-", ".-"], "&&&&", 3),
        # ),
        ), (
            (["&-"], "&&", 3), # ＋
        ),
    )

    MAIN_ROWS_KEYS_STAGGERED_RIGHT: KeyColumnsTuple = (
        (
            (["*"], "🞱", 3),
        ), (
            (["*", "-F"], ""),
            (["*", "-F", "-R"], ""),
            (["*", "-R"], ""),
        ), (
            (["-F"], "F", 1, "6"),
            (["-F", "-R"], ""),
            (["-R"], "R"),
        ), (
            (["-P"], "P", 1, "7"),
            (["-P", "-B"], ""),
            (["-B"], "B"),
        ), (
            (["-L"], "L", 1, "8"),
            (["-L", "-G"], ""),
            (["-G"], "G"),
        ), (
            (["-T"], "T", 1, "9"),
            (["-T", "-S"], ""),
            (["-S"], "S"),
        ), (
            (["-T", "-D"], ""),
            (["-T", "-S", "-D", "-Z"], ""),
            (["-S", "-Z"], ""),
        ), (
            (["-D"], "D"),
            (["-D", "-Z"], ""),
            (["-Z"], "Z"),
        ),
    )

    # [keys], label, (row, column, rowSpan?, columnSpan?), numberLabel?
    MAIN_ROWS_KEYS_GRID: KeyGridTuple = (
        (["^-"], "^", (TOP_ROW, 0)),
        (["+-"], "+", (LOW_ROW, 0)),
        (["^-", "S-"], "", (TOP_ROW, 1)),
        (["+-", "S-"], "", (LOW_ROW, 1)),
        (["S-"], "S", (TOP_ROW, 2, 3, 1), "1"),
        (["T-"], "T", (TOP_ROW, 3), "2"),
        (["K-"], "K", (LOW_ROW, 3)),
        (["P-"], "P", (TOP_ROW, 4), "3"),
        (["W-"], "W", (LOW_ROW, 4)),
        (["H-"], "H", (TOP_ROW, 5), "4"),
        (["R-"], "R", (LOW_ROW, 5)),
        (["H-", "*"], "", (TOP_ROW, 6)),
        (["R-", "*"], "", (LOW_ROW, 6)),
        (["*"], "🞱", (TOP_ROW, 7, 3, 1)),
        (["*", "-F"], "", (TOP_ROW, 8)),
        (["*", "-R"], "", (LOW_ROW, 8)),
        (["-F"], "F", (TOP_ROW, 9), "6"),
        (["-R"], "R", (LOW_ROW, 9)),
        (["-P"], "P", (TOP_ROW, 10), "7"),
        (["-B"], "B", (LOW_ROW, 10)),
        (["-L"], "L", (TOP_ROW, 11), "8"),
        (["-G"], "G", (LOW_ROW, 11)),
        (["-T"], "T", (TOP_ROW, 12), "9"),
        (["-S"], "S", (LOW_ROW, 12)),
        (["-T", "-D"], "", (TOP_ROW, 13)),
        (["-S", "-Z"], "", (LOW_ROW, 13)),
        (["-D"], "D", (TOP_ROW, 14)),
        (["-Z"], "Z", (LOW_ROW, 14)),

        (["^-", "+-"], "", (LOW_COMPOUND_ROW, 0)),
        (["^-", "+-", "S-"], "", (LOW_COMPOUND_ROW, 1)),
        (["T-", "K-"], "", (LOW_COMPOUND_ROW, 3)),
        (["P-", "W-"], "", (LOW_COMPOUND_ROW, 4)),
        (["H-", "R-"], "", (LOW_COMPOUND_ROW, 5)),
        (["H-", "R-", "*"], "", (LOW_COMPOUND_ROW, 6)),
        (["*", "-R", "-F"], "", (LOW_COMPOUND_ROW, 8)),
        (["-F", "-R"], "", (LOW_COMPOUND_ROW, 9)),
        (["-P", "-B"], "", (LOW_COMPOUND_ROW, 10)),
        (["-L", "-G"], "", (LOW_COMPOUND_ROW, 11)),
        (["-T", "-S"], "", (LOW_COMPOUND_ROW, 12)),
        (["-T", "-S", "-D", "-Z"], "", (LOW_COMPOUND_ROW, 13)),
        (["-D", "-Z"], "", (LOW_COMPOUND_ROW, 14)),
    )

    VOWEL_ROW_KEYS_LEFT: KeyGridTuple = (
        (["A-"], "A", (0, 0), "5"),
        (["A-", "O-"], "", (0, 1)),
        (["O-"], "O", (0, 2), "0"),
        
        (["#", "A-"], "", (1, 0)),
        (["#", "A-", "O-"], "", (1, 1)),
        (["#", "O-"], "", (1, 2)),

        (["#"], "#", (2, 0, 1, 3)),
    )

    VOWEL_ROW_KEYS_RIGHT: KeyGridTuple = (
        (["-E"], "E", (0, 0)),
        (["-E", "-U"], "", (0, 1)),
        (["-U"], "U", (0, 2)),
        
        (["_", "-E"], "", (1, 0)),
        (["_", "-E", "-U"], "", (1, 1)),
        (["_", "-U"], "", (1, 2)),

        (["_"], "_", (2, 0, 1, 3)),
    )

    #endregion

    #region Size values

    # in centimeters
    key_width = settings.key_width_ref
    key_height = settings.key_height_ref
    compound_key_size = settings.compound_key_size_ref


    """Computes the size of a key that has half of a compound key cutting into it"""
    reduced_size: Callable[[Ref[float], Ref[float]], Ref[float]] = lambda key_size_ref, compound_size_ref: (
            computed(lambda: key_size_ref.value - compound_size_ref.value / 2,
                    key_size_ref, compound_size_ref)
    )

    reduced_key_width = reduced_size(key_width, compound_key_size)
    reduced_key_height = reduced_size(key_height, compound_key_size)

    """ key_height_num_bar = computed(lambda: key_height.value / 2,
            key_height) """

    index_stretch = settings.index_stretch_ref
    pinky_stretch = settings.pinky_stretch_ref

    vowel_set_offset_fac = settings.vowel_set_offset_fac_ref
    vowel_set_offset = computed(lambda: key_width.value * vowel_set_offset_fac.value,
            key_width, vowel_set_offset_fac)

    """Default row heights for every main rows column"""
    row_heights = (
        reduced_key_height,  # top row
        compound_key_size,
        reduced_key_height,  # bottom row
    )


    # Row heights for specific columns in staggered mode

    compound_height_small = computed(lambda: compound_key_size.value * 0.75,
            compound_key_size)
    reduced_height_small = reduced_size(key_height, compound_height_small)

    row_heights_small = (
        reduced_height_small,
        compound_height_small,
        reduced_height_small,
    )


    row_heights_staggered_left = (
        *(row_heights_small,) * 3,
        *(row_heights,) * 2,
        *(row_heights_small,) * 4,
    )

    row_heights_staggered_right = (
        *(row_heights_small,) * 3,
        *(row_heights,) * 2,
        *(row_heights_small,) * 3,
    )


    index_compound_width = computed(lambda: compound_key_size.value * 0.6,
            compound_key_size)    
    base_index_width = computed(lambda: key_width.value + index_stretch.value,
            key_width, index_stretch)
    reduced_index_width = reduced_size(base_index_width, index_compound_width)


    END_COLUMN_WIDTH_BOOST = 0.4

    end_column_compound_width = computed(lambda: compound_key_size.value * 0.875,
            compound_key_size)
    
    end_column_width = reduced_size(
        computed(lambda: key_width.value + END_COLUMN_WIDTH_BOOST,
                key_width),
        end_column_compound_width,
    )
    inner_end_column_width = reduced_size(
        computed(lambda: key_width.value + pinky_stretch.value,
                key_width, pinky_stretch),
        end_column_compound_width,
    )

    center_key_width = computed(lambda: reduced_key_width.value + key_width.value / 2,
                reduced_key_width, key_width)

    col_widths_staggered_left = (
        end_column_width,
        end_column_compound_width,
        inner_end_column_width,
        key_width,
        key_width,
        reduced_index_width,  # H-, R-
        index_compound_width,
        center_key_width, # &
        # computed(lambda: (key_width.value * 1.5 - compound_key_size.value / 2) / 2,
        #         key_width, compound_key_size),  # +
        # computed(lambda: (key_width.value * 1.5 - compound_key_size.value / 2) / 2,
        #         key_width, compound_key_size),  # .
    )

    col_widths_staggered_right = (
        center_key_width, # *
        index_compound_width,
        reduced_index_width,  # -F, -R
        key_width,
        key_width,
        inner_end_column_width,  # -T, -S
        end_column_compound_width,
        end_column_width,  # -D, -Z
    )

    VOWEL_KEY_WIDTH_BOOST = 0.25

    vowel_set_widths = (
        computed(lambda: reduced_key_width.value + VOWEL_KEY_WIDTH_BOOST,
                reduced_key_width),
        compound_key_size,
        computed(lambda: reduced_key_width.value + VOWEL_KEY_WIDTH_BOOST,
                reduced_key_width),
    )

    vowel_compound_key_height = index_compound_width

    vowel_set_heights = (
        reduced_size(key_height, vowel_compound_key_height),
        vowel_compound_key_height,
        reduced_size(computed(lambda: key_height.value * 0.75, key_height), vowel_compound_key_height),
    )

    index_stagger_fac = settings.index_stagger_fac_ref
    middle_stagger_fac = settings.middle_stagger_fac_ref
    ring_stagger_fac = settings.ring_stagger_fac_ref
    pinky_stagger_fac = settings.pinky_stagger_fac_ref

    index_offset = computed(lambda: key_width.value * index_stagger_fac.value,
            key_width, index_stagger_fac)
    middle_offset = computed(lambda: key_width.value * middle_stagger_fac.value,
            key_width, middle_stagger_fac)
    ring_offset = computed(lambda: key_width.value * ring_stagger_fac.value,
            key_width, ring_stagger_fac)
    pinky_offset = computed(lambda: key_width.value * pinky_stagger_fac.value,
            key_width, pinky_stagger_fac)


    col_offsets_staggered_left = (
        pinky_offset,
        pinky_offset,
        pinky_offset,  # S-
        ring_offset,  # T-, K-
        middle_offset,  # P-, W-
        index_offset,  # H-, R-
        index_offset,
        index_offset,
        index_offset,
    )

    col_offsets_staggered_right = (
        index_offset,
        index_offset,
        index_offset,  # -F, -R
        middle_offset,  # -P, -B
        ring_offset,  # -L, -G
        pinky_offset,  # -T, -S
        pinky_offset,
        pinky_offset, # -D, -Z
    )

    col_widths_grid = (
        *col_widths_staggered_left[:-1],
        computed(lambda: reduced_key_width.value * 2 + key_width.value * 2.5,
                reduced_key_width, key_width),  # *
        *col_widths_staggered_right[1:],
    )

    #endregion


    return LayoutDescriptor(
        MAIN_ROWS_STAGGERED_LEFT=MAIN_ROWS_KEYS_STAGGERED_LEFT,
        MAIN_ROWS_STAGGERED_RIGHT=MAIN_ROWS_KEYS_STAGGERED_RIGHT,
        col_widths_staggered_left=col_widths_staggered_left,
        col_widths_staggered_right=col_widths_staggered_right,
        row_heights_staggered_left=row_heights_staggered_left,
        row_heights_staggered_right=row_heights_staggered_right,
        col_offsets_staggered_left=col_offsets_staggered_left,
        col_offsets_staggered_right=col_offsets_staggered_right,
        TALLEST_COLUMN_INDEX_LEFT=3,
        TALLEST_COLUMN_INDEX_RIGHT=3,

        N_INDEX_COLS_LEFT=5,
        N_INDEX_COLS_RIGHT=5,

        MAIN_ROWS_GRID=MAIN_ROWS_KEYS_GRID,
        row_heights_grid=row_heights,
        col_widths_grid=col_widths_grid,
        ASTERISK_COLUMN_INDEX_GRID=7,

        VOWEL_ROW_KEYS_LEFT=VOWEL_ROW_KEYS_LEFT,
        VOWEL_ROW_KEYS_RIGHT=VOWEL_ROW_KEYS_RIGHT,
        vowel_set_widths=vowel_set_widths,
        vowel_set_heights=vowel_set_heights,
        vowel_set_offset=vowel_set_offset,

        LOW_ROW=LOW_ROW,
    )


from plover.machine.base import ThreadedStenotypeBase

from typing import Any

class NoneMachine(ThreadedStenotypeBase):
    """Machine that does nothing"""

    KEYS_LAYOUT: str = ""

    def __init__(self, params: dict[str, Any]):
        super().__init__()
    
    def run(self):
        self._ready()

    def start_capture(self):
        super().start_capture()

    def stop_capture(self):
        super().stop_capture()

    @classmethod
    def get_option_info(cls):
        return {}


from PyQt5.QtCore import (
    Qt,
    pyqtBoundSignal,
)
from PyQt5.QtWidgets import (
    QWidget,
    QDialog,
    QVBoxLayout,
    QHBoxLayout,
    QGridLayout,
    QGroupBox,
    QButtonGroup,
    QRadioButton,
    QCheckBox,
    QLabel,
    QSizePolicy,
)
from PyQt5.QtGui import (
    QFont,
    QTextBlock,
    QTextDocument,
)

from typing import Callable, TYPE_CHECKING
if TYPE_CHECKING:
    from ..Main import Main
else:
    Main = object

from .FloatInput import FloatSlider, FloatEntry
from ..settings import Settings, KeyLayout
from ..util import Ref, on, watch_many, FONT_FAMILY


class SettingsDialog(QDialog):
    def __init__(self, settings: Settings, parent: Main=None):
        super().__init__(parent)
        
        self.__settings = settings
    
        self.__setup_ui()

    def __setup_ui(self):
        self.setWindowTitle("Touch stenotype settings")

        size = self.size()
        size.setWidth(400)
        self.resize(size)

        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)

        self.setFont(QFont(FONT_FAMILY, 11))


        settings = self.__settings


        key_layout_box = QGroupBox("Key layout", self)
        key_layout_group = QButtonGroup(key_layout_box)

        key_layout_radios = {
            KeyLayout.STAGGERED: QRadioButton("Staggered", key_layout_box),
            KeyLayout.GRID: QRadioButton("Grid", key_layout_box),
        }
        key_layout_radios[settings.key_layout].setChecked(True)

        self.__key_layout_radios = {
            id(button): value
            for value, button in key_layout_radios.items()
        }

        key_layout_box_layout = QVBoxLayout()
        for radio in key_layout_radios.values():
            key_layout_box_layout.addWidget(radio)
            key_layout_group.addButton(radio)

        key_layout_box_layout.addStretch(1)
        key_layout_box.setLayout(key_layout_box_layout)

        @on(key_layout_group.buttonToggled)
        def update_keyboard_layout(button: QRadioButton, checked: bool):
            if not checked: return
            settings.key_layout = self.__key_layout_radios[id(button)]


        stroke_preview_box = QGroupBox("Stroke preview", self)

        stroke_preview_checkboxes = (
            QCheckBox("Show stroke", stroke_preview_box),
            QCheckBox("Show translation", stroke_preview_box),
        )
        stroke_preview_checkboxes[0].setChecked(settings.stroke_preview_stroke)
        stroke_preview_checkboxes[1].setChecked(settings.stroke_preview_translation)

        stroke_preview_box_layout = QVBoxLayout()
        for checkbox in stroke_preview_checkboxes:
            stroke_preview_box_layout.addWidget(checkbox)
            
        stroke_preview_box_layout.addStretch(1)
        stroke_preview_box.setLayout(stroke_preview_box_layout)

        @on(stroke_preview_checkboxes[0].toggled)
        def update_stroke_preview_stroke(checked: bool):
            settings.stroke_preview_stroke = checked

        @on(stroke_preview_checkboxes[1].toggled)
        def update_stroke_preview_translation(checked: bool):
            settings.stroke_preview_translation = checked


        misc_box = QGroupBox("Misc", self)
        misc_box_layout = QGridLayout()

        opacity_entry, opacity_slider = _build_entry_slider_pair(
            settings.window_opacity_ref,
            min=0.25,
            max=1,
            parent=misc_box,
        )
        
        misc_box_layout.addWidget(QLabel("Window opacity"), 0, 0, 1, 2)
        misc_box_layout.addWidget(opacity_slider, 1, 0)
        misc_box_layout.addWidget(opacity_entry, 1, 1)

        misc_box.setLayout(misc_box_layout)



        size_box = QGroupBox("Key and layout geometry", self)
        size_box_layout = QGridLayout()

        for size_box_index, (label, box, slider, after_label) in enumerate((
            ("Base key width",
                *_build_entry_slider_pair(
                    settings.key_width_ref,
                    min=0.5,
                    max=3,
                    spin_box_step=0.05,
                    parent=size_box,
                ), "cm"),
            ("Base key height",
                *_build_entry_slider_pair(
                    settings.key_height_ref,
                    min=0.5,
                    max=3,
                    spin_box_step=0.05,
                    parent=size_box,
                ), "cm"),
            ("Compound key size",
                    *(compound_input_pair := _build_entry_slider_pair(
                        settings.compound_key_size_ref,
                        min=0.25,
                        max=min(settings.key_width, settings.key_height),
                        spin_box_step=0.05,
                        parent=size_box,
                    )), "cm"),
            ("Index finger stretch",
                *_build_entry_slider_pair(
                    settings.index_stretch_ref,
                    min=-0.25,
                    max=1,
                    spin_box_step=0.05,
                    parent=size_box,
                ), "cm"),
            ("Pinky finger stretch",
                *_build_entry_slider_pair(
                    settings.pinky_stretch_ref,
                    min=0,
                    max=1.5,
                    spin_box_step=0.05,
                    parent=size_box,
                ), "cm"),
            ("Vowels offset factor",
                *_build_entry_slider_pair(
                    settings.vowel_set_offset_fac_ref,
                    min=0,
                    max=1.5,
                    spin_box_step=0.05,
                    parent=size_box,
                ), ""),
            ("Main rows angle",
                *_build_entry_slider_pair(
                    settings.main_rows_angle_ref,
                    min=0,
                    max=45,
                    spin_box_step=0.5,
                    parent=size_box,
                ), "°"),
            ("Vowel rows angle",
                *_build_entry_slider_pair(
                    settings.vowel_rows_angle_ref,
                    min=0,
                    max=75,
                    spin_box_step=0.5,
                    parent=size_box,
                ), "°"),
        )):
            size_box_layout.addWidget(QLabel(label), size_box_index * 2, 0, 1, 3)
            size_box_layout.addWidget(slider, size_box_index * 2 + 1, 0)
            size_box_layout.addWidget(box, size_box_index * 2 + 1, 1)
            size_box_layout.addWidget(QLabel(after_label), size_box_index * 2 + 1, 2)


        compound_key_box, compound_key_slider = compound_input_pair
        @watch_many(settings.key_width_ref.change, settings.key_height_ref.change)
        def set_compound_key_size_max():
            new_max = min(settings.key_width, settings.key_height)

            compound_key_box.setMaximum(new_max)
            compound_key_slider.max = new_max
            

        size_box_index += 1

        size_box_layout.addWidget(QLabel("Column stagger factors"), size_box_index * 2, 0, 1, 3)

        stagger_layout = QGridLayout()
        stagger_layout.setContentsMargins(0, 0, 0, 0)
        stagger_layout.setRowMinimumHeight(0, 120)
        for stagger_index, ref in enumerate((
            settings.index_stagger_fac_ref,
            settings.middle_stagger_fac_ref,
            settings.ring_stagger_fac_ref,
            settings.pinky_stagger_fac_ref,
        )):
            entry, slider = _build_entry_slider_pair(
                ref,
                min=0,
                max=2,
                spin_box_step=0.05,
                slider_orientation=Qt.Vertical,
                parent=size_box,
            )

            stagger_layout.addWidget(slider, 0, stagger_index, Qt.AlignHCenter)
            stagger_layout.addWidget(entry, 1, stagger_index, Qt.AlignHCenter)

        size_box_layout.addLayout(stagger_layout, size_box_index * 2 + 1, 0, 1, 3)
        
        # size_box_layout.setRowStretch(2 * (size_box_index + 1), 1)
        size_box.setLayout(size_box_layout)


        label_resizing = QLabel("Resize the stenotype window to adjust spacing", self)
        label_resizing.setWordWrap(True)
        label_resizing.setStyleSheet("font-style: italic; color: #7f000000;")

        label_troubleshooting = QLabel("If there are issues with responsiveness, check the plugin description (§ Additional setup) for possible solutions",
                self)
        label_troubleshooting.setWordWrap(True)
        label_troubleshooting.setStyleSheet("font-style: italic; color: #7f000000;")


        layout = QGridLayout()
        layout.addWidget(key_layout_box, 0, 0)
        layout.addWidget(stroke_preview_box, 1, 0)
        layout.addWidget(misc_box, 2, 0)
        layout.addWidget(size_box, 0, 1, 4, 1)
        layout.setRowStretch(3, 1)
        layout.addWidget(label_resizing, 4, 0, 1, 2)
        layout.addWidget(label_troubleshooting, 5, 0, 1, 2)
        # layout.addWidget(sizes_box)
        self.setLayout(layout)



def _build_entry_slider_pair(
    ref: Ref[float],
    min: float=0,
    max: float=1,
    spin_box_step=0.1,
    slider_orientation: Qt.Orientation=Qt.Horizontal,
    parent: QWidget=None,
) -> tuple[FloatEntry, FloatSlider]:
    entry = FloatEntry(ref.value,
            min=min,
            max=max,
            spin_step=spin_box_step,
            parent=parent)
    slider = FloatSlider(ref.value,
            min=min,
            max=max,
            orientation=slider_orientation,
            parent=parent)

    last_edit_from_entry = False
    last_edit_from_slider = False

    @on(entry.input)
    def on_entry_input(value: float):
        nonlocal last_edit_from_entry
        last_edit_from_entry = True
    @on(slider.input)
    def on_slider_input(value: float):
        nonlocal last_edit_from_slider
        last_edit_from_slider = True

    @on(entry.input)
    @on(slider.input)
    def update_settings(value: float):
        ref.set(value)

    @on(ref.change)
    def update_displays(value: float):
        nonlocal last_edit_from_entry
        nonlocal last_edit_from_slider

        # if not box.hasFocus():
        if not last_edit_from_entry:
            entry.current_value = value
        # if not slider.hasFocus():
        if not last_edit_from_slider:
            slider.current_value = value

        last_edit_from_entry = False
        last_edit_from_slider = False

    return entry, slider



from PyQt5.QtCore import (
    Qt,
    QEvent,
    pyqtProperty,
)
from PyQt5.QtWidgets import (
    QWidget,
    QToolButton,
    QSizePolicy,
)
from PyQt5.QtGui import (
    QFont,
)


from plover_touchscreen_stenotype.util import FONT_FAMILY


class KeyWidget(QToolButton):
    #region Overrides

    def __init__(self, values: list[str], label: str, parent: QWidget=None):
        # super().__init__(label, parent)
        super().__init__(parent)

        self.values = values

        self.__touched = False
        self.__matched = False

        self.__setup_ui(label)

    def event(self, event: QEvent):
        # Prevents automatic button highlighting
        if event.type() == QEvent.HoverEnter:
            self.setAttribute(Qt.WA_UnderMouse, False)

        return super().event(event)

    #endregion

    def __setup_ui(self, label):
        self.setText(label)

        if label:
            self.setFont(QFont(FONT_FAMILY, 16))


        # self.setMinimumSize(0, 0)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        # self.setAttribute(Qt.WA_AcceptTouchEvents)
        self.setFocusPolicy(Qt.NoFocus)


    def num_bar_pressed_handler(self, label: str, num_bar_label: str):
        return lambda num_bar_pressed: self.setText(num_bar_label if num_bar_pressed else label)


    @pyqtProperty(bool)
    def touched(self):
        return self.__touched

    @touched.setter
    def touched(self, touched: bool):
        self.__touched = touched

    @pyqtProperty(bool)
    def matched(self):
        return self.__matched

    @matched.setter
    def matched(self, matched: bool):
        self.__matched = matched


from PyQt5.QtCore import (
    QObject,
    QTimer,
    pyqtSignal,
    pyqtBoundSignal,
)
from PyQt5.QtWidgets import (
    QWidget,
)
from PyQt5.QtGui import (
    QScreen,
)

from functools import partial
from dataclasses import dataclass
from typing import TypeVar, Generic, Any, Callable, Iterable, cast


class UseDpi(QObject):
    """Composable that handles DPI-responsivity."""

    change = pyqtSignal()
    change_logical = pyqtSignal()

    def __init__(self, widget: QWidget):
        super().__init__(widget)

        self.__widget = widget
        self.__current_screen = widget.screen()

        self.__current_screen.physicalDotsPerInchChanged.connect(self.__on_screen_physcial_dpi_change)
        self.__current_screen.logicalDotsPerInchChanged.connect(self.__on_screen_logical_dpi_change)
        # `widget.window().windowHandle()` may initially be None on Linux
        QTimer.singleShot(0,
            lambda: widget.window().windowHandle().screenChanged.connect(self.__on_screen_change)
        )

    def cm(self, cm: float) -> int:
        """Converts cm to px using the current physical DPI."""
        return round(cm * self.__widget.screen().physicalDotsPerInch() / 2.54)

    def dp(self, dp: float) -> int:
        """Converts dp to px using the current physical DPI. (Defines dp as the length of a pixel on a 96 dpi screen.)"""
        return round(dp * self.__widget.screen().physicalDotsPerInch() / 96)

    def pt(self, pt: float) -> int:
        """Converts pt to px using the current logical DPI."""
        return round(pt * self.__widget.screen().logicalDotsPerInch() / 72)

    def __on_screen_change(self, screen: QScreen):
        self.__current_screen.physicalDotsPerInchChanged.disconnect(self.__on_screen_physcial_dpi_change)
        self.__current_screen.logicalDotsPerInchChanged.disconnect(self.__on_screen_logical_dpi_change)

        self.__current_screen = screen
        screen.physicalDotsPerInchChanged.connect(self.__on_screen_physcial_dpi_change)
        screen.logicalDotsPerInchChanged.connect(self.__on_screen_logical_dpi_change)

        self.change.emit()
        self.change_logical.emit()

    def __on_screen_physcial_dpi_change(self, dpi: float):
        self.change.emit()

    def __on_screen_logical_dpi_change(self, dpi: float):
        self.change_logical.emit()


T = TypeVar("T")
class RefAttr(Generic[T]):
    """Descriptor for one shallow reactive value."""

    def __init__(self, expected_type: type[T]):
        # self.signal = pyqtSignal(expected_type)
        pass

    def __set_name__(self, owner_class: type, attr_name: str):
        self.__ref_name = f"{attr_name}_ref"
        # self.__signal_name = f"{attr_name}_change"

    def __get__(self, instance: Any, owner_class: type) -> T:
        if instance is None:
            return self
        return cast(Ref[T], getattr(instance, self.__ref_name)).value

    def __set__(self, instance: Any, value: T):
        if not hasattr(instance, self.__ref_name):
            setattr(instance, self.__ref_name, Ref(value))
        else:
            cast(Ref[T], getattr(instance, self.__ref_name)).value = value

        # setattr(instance, self.__private_attr_name, value)
        # cast(pyqtBoundSignal, getattr(instance, self.__signal_name)).emit(value)

    def ref_getter(self) -> "_RefGetter[T]":
        return _RefGetter()


class _RefGetter(Generic[T]):
    def __set_name__(self, owner_class: type, attr_name: str):
        self.__private_attr_name = f"__{owner_class.__name__}_{attr_name}"

    def __get__(self, instance: Any, owner_class: type) -> "Ref[T]":
        if instance is None:
            return self
        return getattr(instance, self.__private_attr_name)

    def __set__(self, instance: Any, value: "Ref[T]"):
        setattr(instance, self.__private_attr_name, value)


class Ref(QObject, Generic[T]):
    change = pyqtSignal(object) # T

    def __init__(self, value: T):
        super().__init__()
        self.__value = value

    @property
    def value(self) -> T:
        return self.__value

    @value.setter
    def value(self, value: T):
        self.__value = value
        self.change.emit(value)

    def set(self, value: T):
        """Alias for `value` setter"""
        # so values can be set from lambdas
        self.value = value


def _create_computed(handler: Callable[[], T]):
    ref = Ref(handler())

    def recompute_value():
        # Blocking signals prevents a computed ref from recalculating multiple times due to simulataneous changes
        if ref.signalsBlocked(): return

        ref.value = handler()

        ref.blockSignals(True)
        QTimer.singleShot(0, lambda: ref.blockSignals(False))

    return ref, recompute_value


def computed(handler: Callable[[], T], *dependency_refs: Ref[T]):
    ref, recompute = _create_computed(handler)
    for dependency in dependency_refs:
        dependency.change.connect(recompute)

    return ref


def computed_on_signal(handler: Callable[[], T], dependency_signal: pyqtBoundSignal):
    ref, recompute = _create_computed(handler)
    _connect(dependency_signal, recompute, parent=ref)

    return ref


# TODO mix between synchrony and asynchrony below may lead to inconsistent/unexpected behavior

def _connect(signal: pyqtBoundSignal, handler: Callable[..., None], parent: "QObject | None"=None):
    # Disconnecting using `connection` instead of `handler` allows errors to be caught properly when attempting to
    # disconnect after the signal parents have been destroyed (why?)
    connection = signal.connect(handler)

    if parent is not None:
        @on(parent.destroyed)
        def disconnect():
            # try-except might not be ideal
            try:
                signal.disconnect(connection)
            except TypeError:
                pass

    return handler


def _connect_many(signals: Iterable[pyqtBoundSignal], handler: Callable[..., None], parent: "QObject | None"=None):
    # Blocking prevents handler from firing multiple times due to simulataneous changes
    # Handler is contained in QTimer.singleShot callback so that all watched signals will have been run
    # TODO ^ not guaranteed? (eg, if one watcher is invoked because of another)
    waiting = False

    def call_later(*args: Any):
        nonlocal waiting

        if waiting: return

        waiting = True
        QTimer.singleShot(0, call_and_unblock)

    def call_and_unblock():
        nonlocal waiting

        handler()
        waiting = False


    connections = tuple(signal.connect(call_later) for signal in signals)

    if parent is not None:
        @on(parent.destroyed)
        def disconnect_all():
            for signal, connection in zip(signals, connections):
                try:
                    signal.disconnect(connection)
                except TypeError:
                    pass

    return handler


def on(signal: pyqtBoundSignal, parent: "QObject | None"=None):
    """Decorator factory. Connects a function to a signal.

        :param parent: QObject that, when destroyed, will cause the handler to be disconnected from the signal.
    """

    return partial(_connect, signal, parent=parent)


def on_many(*signals: pyqtBoundSignal, parent: "QObject | None"=None):
    """Decorator factory. Connects a function to an arbitrary number of signals.

        :param parent: QObject that, when destroyed, will cause the handler to be disconnected from all given signals.
    """

    return partial(_connect_many, signals, parent=parent)


def watch(signal: pyqtBoundSignal, parent: "QObject | None"=None):
    """Decorator factory. Calls a function immediately and connects it to a signal.

        :param parent: QObject that, when destroyed, will cause the handler to be disconnected from the signal.
    """

    def run_and_connect(handler: Callable[..., None]):
        handler()
        return _connect(signal, handler, parent=parent)

    return run_and_connect


def watch_many(*signals: pyqtBoundSignal, parent: "QObject | None"=None):
    """Decorator factory. Calls a function immediately and connects it to an arbitrary number of signals.

        :param parent: QObject that, when destroyed, will cause the handler to be disconnected from all given signals.
    """

    def run_and_connect(handler: Callable[..., None]):
        handler()
        return _connect_many(signals, handler, parent=parent)

    return run_and_connect


FONT_FAMILY = "Atkinson Hyperlegible, Segoe UI, Ubuntu"

KEY_STYLESHEET = """
KeyWidget[matched="true"] {
    background: #6f9f86;
    color: #fff;
    border: 1px solid;
    border-color: #2a6361 #2a6361 #1f5153 #2a6361;
}

KeyWidget[touched="true"] {
    background: #41796a;
}
"""


KeyColumnsTuple = tuple[
    tuple[
        tuple[list[str], str, "int | None", "str | None"]
    ]
]
KeyGridTuple = tuple[
    tuple[
        list[str],
        str,
        "tuple[int, int] | tuple[int, int, int, int]",
        "str | None",
    ]
]
SizeTuple = tuple[Ref[float]]

@dataclass
class LayoutDescriptor:
    MAIN_ROWS_STAGGERED_LEFT: KeyColumnsTuple
    MAIN_ROWS_STAGGERED_RIGHT: KeyColumnsTuple
    col_widths_staggered_left: SizeTuple
    col_widths_staggered_right: SizeTuple
    col_offsets_staggered_left: SizeTuple
    col_offsets_staggered_right: SizeTuple
    row_heights_staggered_left: SizeTuple
    row_heights_staggered_right: SizeTuple
    TALLEST_COLUMN_INDEX_LEFT: float
    TALLEST_COLUMN_INDEX_RIGHT: float

    N_INDEX_COLS_LEFT: float
    N_INDEX_COLS_RIGHT: float

    MAIN_ROWS_GRID: KeyGridTuple
    row_heights_grid: SizeTuple
    col_widths_grid: SizeTuple
    ASTERISK_COLUMN_INDEX_GRID: float

    VOWEL_ROW_KEYS_LEFT: KeyGridTuple
    VOWEL_ROW_KEYS_RIGHT: KeyGridTuple
    vowel_set_widths: SizeTuple
    vowel_set_heights: SizeTuple
    vowel_set_offset: Ref[float]

    LOW_ROW: float


from plover.gui_qt.tool import Tool
from plover.gui_qt.utils import ToolBar
from plover.gui_qt.engine import Engine
from plover.steno import Stroke
from plover.oslayer import PLATFORM

from PyQt5.QtCore import (
    Qt,
    QSize,
    QSettings,
)
from PyQt5.QtWidgets import (
    QWidget,
    QGridLayout,
    QAction,
)
from PyQt5.QtGui import (
    QIcon,
)


from .settings import Settings
from .util import Ref, watch
from .widgets.KeyboardWidget import KeyboardWidget
from .widgets.StrokePreview import StrokePreview
from .widgets.SettingsDialog import SettingsDialog


class Main(Tool):
    #region Overrides

    TITLE = "Touchscreen stenotype"
    ICON = ""
    ROLE = "touchscreen_stenotype"

    def __init__(self, engine: Engine):
        super().__init__(engine)

        self.engine = engine # Override for type hint
        self.__last_stroke_from_widget = False
        """Whether the last emitted stroke originated from this Tool"""
        self.__last_stroke_keys: set[str] | None = None
        self.__last_stroke_engine_enabled = False

        self.__settings = Settings()
        self.restore_state()
        self.finished.connect(self.save_state)
        self.__setup_ui()

        engine.signal_stroked.connect(self.__on_stroked)


    def _restore_state(self, settings: QSettings):
        self.__settings.load(settings)

    def _save_state(self, settings: QSettings):
        self.__settings.save(settings)

    #endregion

    def __setup_ui(self):
        self.setAttribute(Qt.WA_AcceptTouchEvents)
        # self.setAttribute(Qt.WA_ShowWithoutActivating)
        # self.setFocusPolicy(Qt.NoFocus)

        # https://stackoverflow.com/questions/71084136/how-to-set-focus-to-the-old-window-on-button-click-in-pyqt5-python
        # self.setWindowFlags(Qt.Dialog | Qt.WindowStaysOnTopHint | Qt.BypassWindowManagerHint | Qt.WindowDoesNotAcceptFocus)
        self.setWindowFlags(Qt.WindowStaysOnTopHint | Qt.WindowMinimizeButtonHint | Qt.WindowCloseButtonHint)

        # For some reason (tested on Windows only), this has to be called before KeyboardWidget is created. Other attributes
        # must also be set before this, otherwise the window will steal focus again
        self.__prevent_window_focus()


        left_right_width_diff = Ref(0)

        self.stroke_preview = stroke_preview = StrokePreview(self.engine, self.__settings, left_right_width_diff, self)

        stenotype = KeyboardWidget(self.__settings, left_right_width_diff, self)
        stenotype.end_stroke.connect(self.__on_stenotype_input)
        stenotype.current_stroke_change.connect(self.__on_stroke_change)

        settings_action = QAction(self)
        settings_action.setText("Settings")
        settings_action.setIcon(QIcon(r":/settings.svg")) # Loads from Plover's application-wide application resources
        settings_action.triggered.connect(self.__launch_settings_dialog)

        toolbar = ToolBar(settings_action)
        toolbar.setFocusPolicy(Qt.NoFocus)
        for button in toolbar.findChildren(QWidget):
            button.setFocusPolicy(Qt.NoFocus)

        toolbar.setIconSize(QSize(48, 48))


        layout = QGridLayout(self)
        layout.addWidget(stroke_preview, 0, 0)
        layout.addWidget(toolbar, 0, 0, Qt.AlignBottom | Qt.AlignLeft)
        layout.addWidget(stenotype, 0, 0)
        self.setLayout(layout)


        @watch(self.__settings.window_opacity_ref.change)
        def set_window_opacity():
            self.setWindowOpacity(self.__settings.window_opacity)


    # https://stackoverflow.com/questions/24582525/how-to-show-clickable-qframe-without-loosing-focus-from-main-window
    # https://stackoverflow.com/questions/68276479/how-to-use-setwindowlongptr-hwnd-gwl-exstyle-ws-ex-noactivate

    # https://docs.microsoft.com/en-us/windows/win32/api/winuser/nf-winuser-setwindowlongptrw
    def __prevent_window_focus(self):
        """
        Prevents the stenotype window from taking focus from other programs when the keys are touched. This is a
        cross-platform polyfill for the window flag `Qt.WindowDoesNotAcceptFocus` (which is nonfunctional on
        Windows).

        See https://bugreports.qt.io/browse/QTBUG-36230, https://forum.qt.io/topic/82493/the-windowdoesnotacceptfocus-flag-is-making-me-thirsty/7 for bug information.
        """

        # On Windows: does not work by itself, but prevents focus from being taken from other Plover windows
        self.setWindowFlag(Qt.WindowDoesNotAcceptFocus)

        if PLATFORM == "win":
            import ctypes
            from ctypes.wintypes import HWND
            import win32con

            window_handle = HWND(int(self.winId()))

            user32 = ctypes.windll.user32
            user32.SetWindowLongPtrW(
                window_handle,
                win32con.GWL_EXSTYLE,
                user32.GetWindowLongPtrW(window_handle, win32con.GWL_EXSTYLE) | win32con.WS_EX_NOACTIVATE | win32con.WS_EX_APPWINDOW,
            )

        # elif PLATFORM == "linux":
        #     # self.setAttribute(Qt.WA_X11DoNotAcceptFocus)


    def __on_stenotype_input(self, stroke_keys: set[str]):
        # Temporarily enable steno output (if not already waiting for a `stroked` hook dispatch)
        if not self.__last_stroke_from_widget:
            self.__last_stroke_engine_enabled = self.engine.output
            self.engine.output = True

        self.__last_stroke_from_widget = True
        self.__last_stroke_keys = stroke_keys
        self.engine._machine._notify(list(stroke_keys))

        # Wait until `stroked` hook is dispatched to reset `self.engine.output`, since it must be True for Suggestions to be shown

        # TODO The current implementation is not infallible because the `stroked` handler does not verify that the stroke it
        # received is the same stroke sent from this method. Multiple strokes may also be sent before the handler is called
        # (can use deque to resolve this)

        self.stroke_preview.finish_stroke()

    def __on_stroked(self, stroke: Stroke):
        if not self.__last_stroke_from_widget or self.__last_stroke_keys != set(stroke.keys()): return

        if self.engine.output == False:
            # The engine output was disabled before it was set again; {PLOVER:TOGGLE} was likely triggered
            self.engine.output = not self.__last_stroke_engine_enabled
        else:
            self.engine.output = self.__last_stroke_engine_enabled
        
        self.__last_stroke_from_widget = False
        self.__last_stroke_keys = None


    def __on_stroke_change(self, stroke_keys: set[str]):
        self.stroke_preview.display_keys(stroke_keys)


    """ def resize_from_center(self, width: int, height: int):
        rect = self.geometry()
        old_center = rect.center()

        # In practice, the new size will be constrained to the minimum size, so it is not necessarily the given size
        new_size = QSize(
            max(self.minimumWidth(), width),
            max(self.minimumHeight(), height),
        )

        rect.setSize(new_size)
        rect.moveCenter(old_center)
        
        self.setGeometry(rect) """


    def __launch_settings_dialog(self):
        dialog = SettingsDialog(self.__settings, self)
        dialog.open()

# def command_open_window(engine: Engine, arg: str):
#     new_window = Main(engine)
#     new_window.show()



from PyQt5.QtCore import (
    QTimer,
)
from PyQt5.QtWidgets import (
    QLayout,
    QGridLayout,
    QHBoxLayout,
    QVBoxLayout,
    QSpacerItem,
    QWidgetItem,
    QLayoutItem,
)

from math import sin, cos, radians
from functools import partial
from typing import TYPE_CHECKING, Callable, Iterable
if TYPE_CHECKING:
    from .KeyboardWidget import KeyboardWidget
else:
    KeyboardWidget = object

from .KeyWidget import KeyWidget
from .RotatableKeyContainer import RotatableKeyContainer
from ..settings import Settings, KeyLayout
from ..util import UseDpi, Ref, computed, on, on_many, watch, watch_many


#region Exports

def use_build_keyboard(settings: Settings, keyboard_widget: KeyboardWidget, dpi: UseDpi):
    from .build_keyboard_config.stenotype_extended_custom import build_layout_descriptor
    layout_descriptor = build_layout_descriptor(settings)

    MAIN_ROWS_STAGGERED_LEFT = layout_descriptor.MAIN_ROWS_STAGGERED_LEFT
    MAIN_ROWS_STAGGERED_RIGHT = layout_descriptor.MAIN_ROWS_STAGGERED_RIGHT
    col_widths_staggered_left = layout_descriptor.col_widths_staggered_left
    col_widths_staggered_right = layout_descriptor.col_widths_staggered_right
    row_heights_staggered_left = layout_descriptor.row_heights_staggered_left
    row_heights_staggered_right = layout_descriptor.row_heights_staggered_right
    col_offsets_staggered_left = layout_descriptor.col_offsets_staggered_left
    col_offsets_staggered_right = layout_descriptor.col_offsets_staggered_right
    TALLEST_COLUMN_INDEX_LEFT = layout_descriptor.TALLEST_COLUMN_INDEX_LEFT
    TALLEST_COLUMN_INDEX_RIGHT = layout_descriptor.TALLEST_COLUMN_INDEX_RIGHT

    N_INDEX_COLS_LEFT = layout_descriptor.N_INDEX_COLS_LEFT
    N_INDEX_COLS_RIGHT = layout_descriptor.N_INDEX_COLS_RIGHT

    MAIN_ROWS_GRID = layout_descriptor.MAIN_ROWS_GRID
    row_heights_grid = layout_descriptor.row_heights_grid
    col_widths_grid = layout_descriptor.col_widths_grid
    ASTERISK_COLUMN_INDEX_GRID = layout_descriptor.ASTERISK_COLUMN_INDEX_GRID

    VOWEL_ROW_KEYS_LEFT = layout_descriptor.VOWEL_ROW_KEYS_LEFT
    VOWEL_ROW_KEYS_RIGHT = layout_descriptor.VOWEL_ROW_KEYS_RIGHT
    vowel_set_widths = layout_descriptor.vowel_set_widths
    vowel_set_heights = layout_descriptor.vowel_set_heights
    vowel_set_offset = layout_descriptor.vowel_set_offset
    
    LOW_ROW = layout_descriptor.LOW_ROW

    
    key_width = settings.key_width_ref
    key_height = settings.key_height_ref
    compound_key_size = settings.compound_key_size_ref

    index_stagger_fac = settings.index_stagger_fac_ref
    middle_stagger_fac = settings.middle_stagger_fac_ref
    ring_stagger_fac = settings.ring_stagger_fac_ref
    pinky_stagger_fac = settings.pinky_stagger_fac_ref

    index_stretch = settings.index_stretch_ref
    pinky_stretch = settings.pinky_stretch_ref

    # in degrees
    main_rows_angle = settings.main_rows_angle_ref
    vowel_rows_angle = settings.vowel_rows_angle_ref
    

    INITIAL_ROWS_GAP_HEIGHT = 2.25


    def build_main_rows_hand_staggered(
        key_widgets: list[KeyWidget],
        keys: tuple[list[str], str, int, str],
        col_widths: tuple[Ref[float]],
        col_offsets: tuple[Ref[float]],
        row_heights: tuple[Ref[float]],
    ) -> QLayout:
        # Parameter defaults on inner functions are used to create closures

        layout = QHBoxLayout()
        for column, col_width_cm, col_offset_cm, heights in zip(keys, col_widths, col_offsets, row_heights):
            column_layout = QVBoxLayout()
            column_layout.setSizeConstraint(QLayout.SetFixedSize)
            column_layout.addStretch(1)

            row_pos = 0

            for values, label, *rest in column:
                row_span: int = rest[0] if len(rest) > 0 else 1
                row_heights_cm = heights[row_pos:row_pos + row_span]


                key_widget = KeyWidget(values, label, keyboard_widget)
                key_widgets.append(key_widget)

                if len(rest) > 1:
                    num_bar_label: str = rest[1]
                    keyboard_widget.num_bar_pressed_change.connect(key_widget.num_bar_pressed_handler(label, num_bar_label))


                if row_pos <= LOW_ROW < row_pos + row_span:
                    def height_boost(col_offset_cm: Ref[float]=col_offset_cm):
                        return col_offset_cm.value / 2
                    row_heights_cm = (*row_heights_cm, computed(height_boost, col_offset_cm))
                        
                @watch_many(dpi.change, col_width_cm.change, *(height.change for height in row_heights_cm), parent=key_widget)
                def resize(
                    key_widget: KeyWidget=key_widget,
                    col_width_cm: Ref[float]=col_width_cm,
                    row_heights_cm: tuple[Ref[float]]=row_heights_cm,
                ):
                    key_widget.setFixedSize(dpi.cm(col_width_cm.value), sum(dpi.cm(height.value) for height in row_heights_cm))


                column_layout.addWidget(key_widget)

                row_pos += row_span


            column_layout.addSpacerItem(column_spacer := QSpacerItem(0, 0))

            @watch_many(dpi.change, col_offset_cm.change, parent=column_layout)
            def resize_column_spacing(
                column_spacer: QSpacerItem=column_spacer,
                col_offset_cm: Ref[float]=col_offset_cm,
            ):
                column_spacer.changeSize(0, dpi.cm(col_offset_cm.value / 2))

        
            layout.addLayout(column_layout)
        layout.setSizeConstraint(QLayout.SetFixedSize)
            
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        return layout


    def build_main_rows_hand_container(
        key_widgets: list[KeyWidget],
        keys: tuple[list[str], str, int, str],
        col_widths: tuple[Ref[float]],
        col_offsets: tuple[float],
        row_heights_by_col: tuple[Ref[float]],
        angle: Ref[float],
        align_left: bool,
    ) -> QLayout:
        layout = build_main_rows_hand_staggered(key_widgets, keys, col_widths, col_offsets, row_heights_by_col)
        container = RotatableKeyContainer.of_layout(layout, align_left, angle.value, keyboard_widget)

        @on(angle.change)
        def update_vowel_angle(value: float):
            container.angle = angle.value


        offset_layout = QVBoxLayout()
        offset_layout.addStretch(1) # Forces containers to be aligned to bottom

        # Allows height to be subtracted from container without being consumed by the stretch spacer
        direct_layout = QVBoxLayout()
        direct_layout.addSpacerItem(offset_spacer := QSpacerItem(0, 0))


        @watch_many(dpi.change, key_width.change, key_height.change, compound_key_size.change,
                pinky_stretch.change, index_stretch.change, index_stagger_fac.change,
                middle_stagger_fac.change, ring_stagger_fac.change, pinky_stagger_fac.change,
                angle.change, parent=container)
        def set_proxy_transform_and_container_size():
            # Removes the additional height caused by the rotation when the layout and key widget corners do not match
            if align_left:
                widths = col_widths_staggered_right[-TALLEST_COLUMN_INDEX_RIGHT:]
            else:
                widths = col_widths_staggered_left[:TALLEST_COLUMN_INDEX_LEFT]

            radius = sum(dpi.cm(width.value) for width in widths)
            offset = radius * -sin(radians(abs(angle.value)))

            offset_spacer.changeSize(0, round(offset))

            container.update_size()

        direct_layout.addWidget(container)

        offset_layout.addLayout(direct_layout)

        return offset_layout


    def build_main_rows_layout_staggered(key_widgets: list[KeyWidget]) -> QLayout:
        layout = QHBoxLayout()
        layout.setSpacing(0)

        layout.addLayout(build_main_rows_hand_container(
                key_widgets, MAIN_ROWS_STAGGERED_LEFT, col_widths_staggered_left, col_offsets_staggered_left,
                row_heights_staggered_left, main_rows_angle, False))
        layout.addStretch(1)
        layout.addLayout(build_main_rows_hand_container(
                key_widgets, MAIN_ROWS_STAGGERED_RIGHT, col_widths_staggered_right, col_offsets_staggered_right,
                row_heights_staggered_right, computed(lambda: -main_rows_angle.value, main_rows_angle), True))

        return layout

        """ container_layout = QVBoxLayout()
        container_layout.addLayout(layout)
        container_layout.addStretch(1)

        return container_layout """


    def build_main_rows_layout_grid(key_widgets: list[KeyWidget]) -> QLayout:
        layout = QGridLayout()
        for values, label, grid_position, *rest in MAIN_ROWS_GRID:
            key_widget = KeyWidget(values, label, keyboard_widget)
            key_widgets.append(key_widget)
            
            if len(rest) > 0:
                num_bar_label: str = rest[0]
                keyboard_widget.num_bar_pressed_change.connect(key_widget.num_bar_pressed_handler(label, num_bar_label))

            layout.addWidget(key_widget, *grid_position)

        for i, size_cm in enumerate(row_heights_grid):
            layout.setRowMinimumHeight(i, dpi.cm(size_cm.value))
            layout.setRowStretch(i, 0)

        for i, size_cm in enumerate(col_widths_grid):
            layout.setColumnMinimumWidth(i, dpi.cm(size_cm.value))
            layout.setColumnStretch(i, 0)


        layout.setColumnStretch(ASTERISK_COLUMN_INDEX_GRID, 1)
        @watch(dpi.change, parent=layout)
        def resize_asterisk_column():
            QTimer.singleShot(0, lambda: layout.setColumnMinimumWidth(ASTERISK_COLUMN_INDEX_GRID, 0))


        @on_many(dpi.change, *(height.change for height in row_heights_grid), *(width.change for width in col_widths_grid), parent=layout)
        def resize_columns():
            for i, size_cm in enumerate(row_heights_grid):
                layout.setRowMinimumHeight(i, dpi.cm(size_cm.value))

            for i, size_cm in enumerate(col_widths_grid):
                if i == ASTERISK_COLUMN_INDEX_GRID: continue
                layout.setColumnMinimumWidth(i, dpi.cm(size_cm.value))


        layout.setSpacing(0)


        return layout


    def build_vowel_set_layout(vowel_key_descriptors: tuple[Iterable[str], str, tuple[int], str], key_widgets: list[KeyWidget]):
        # Parameter defaults on inner functions are used to create closures

        set_layout = QGridLayout()
        set_layout.setSpacing(0)
        set_layout.setContentsMargins(0, 0, 0, 0)
        set_layout.setSizeConstraint(QLayout.SetFixedSize)

        for values, label, grid_position, *rest in vowel_key_descriptors:
            key_widget = KeyWidget(values, label, keyboard_widget)
            key_widgets.append(key_widget)


            if len(grid_position) == 4:
                row_start, col_start, row_span, col_span = grid_position
            elif len(grid_position) == 2:
                row_start, col_start, row_span, col_span = (*grid_position, 1, 1)
            else:
                raise Exception
            
            widths = vowel_set_widths[col_start:col_start + col_span]
            heights = vowel_set_heights[row_start:row_start + row_span]

            if len(rest) > 0:
                num_bar_label: str = rest[0]
                keyboard_widget.num_bar_pressed_change.connect(key_widget.num_bar_pressed_handler(label, num_bar_label))


            @watch_many(dpi.change, *(width.change for width in widths), *(height.change for height in heights), parent=key_widget)
            def resize(
                key_widget: KeyWidget=key_widget,
                widths: tuple[Ref[float]]=widths,
                heights: tuple[Ref[float]]=heights,
            ):
                key_widget.setFixedSize(
                    sum(dpi.cm(width_cm.value) for width_cm in widths),
                    sum(dpi.cm(height_cm.value) for height_cm in heights),
                )


            set_layout.addWidget(key_widget, *grid_position)

        return set_layout

    def build_vowel_row_containing_sets(left_set: QLayoutItem, right_set: QLayoutItem) -> QLayout:
        layout = QHBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)

        layout.addSpacerItem(left_spacer := QSpacerItem(0, 0))
        layout.addItem(left_set)
        layout.addStretch(1)
        layout.addItem(right_set)
        layout.addSpacerItem(right_spacer := QSpacerItem(0, 0))

        @watch_many(dpi.change, key_width.change, pinky_stretch.change, index_stretch.change, vowel_set_offset.change,
                main_rows_angle.change, *(height.change for height in row_heights_grid), parent=layout)
        def resize_spacing():
            left_bank_width = (
                sum(dpi.cm(col_width.value) for col_width in col_widths_staggered_left[:N_INDEX_COLS_LEFT])
                + dpi.cm(key_width.value)
                + dpi.cm(pinky_stretch.value)
                + dpi.cm(index_stretch.value)
            ) * cos(radians(main_rows_angle.value))
            right_bank_width = (
                sum(dpi.cm(col_width.value) for col_width in col_widths_staggered_right[-N_INDEX_COLS_RIGHT:])
                + dpi.cm(key_width.value)
                + dpi.cm(pinky_stretch.value)
                + dpi.cm(index_stretch.value)
            ) * cos(radians(main_rows_angle.value))

            offset = (dpi.cm(key_width.value) + dpi.cm(vowel_set_offset.value)) * cos(radians(main_rows_angle.value))

            left_bank_spacing = left_bank_width - offset
            right_bank_spacing = right_bank_width - offset

            left_spacer.changeSize(left_bank_spacing, 0)
            right_spacer.changeSize(right_bank_spacing, 0)

        @on_many(key_width.change, vowel_set_offset.change, parent=layout)
        def invalidate_layout():
            layout.invalidate()

        return layout


    def build_vowel_row_layout_staggered(key_widgets: list[KeyWidget]) -> QLayout:
        left_set_layout = build_vowel_set_layout(VOWEL_ROW_KEYS_LEFT, key_widgets)
        left_container = RotatableKeyContainer.of_layout(left_set_layout, False, vowel_rows_angle.value, keyboard_widget)
        
        right_set_layout = build_vowel_set_layout(VOWEL_ROW_KEYS_RIGHT, key_widgets)
        right_container = RotatableKeyContainer.of_layout(right_set_layout, True, -vowel_rows_angle.value, keyboard_widget)

        @on(vowel_rows_angle.change)
        def change_vowel_angle(value: float):
            left_container.angle = value
            right_container.angle = -value

        layout = build_vowel_row_containing_sets(QWidgetItem(left_container), QWidgetItem(right_container))


        @watch_many(dpi.change, key_width.change, key_height.change, compound_key_size.change,
                pinky_stretch.change, index_stretch.change,
                parent=layout)
        def update_sizes():
            left_container.update_size()
            right_container.update_size()
        
        return layout


    def build_vowel_row_layout_grid(key_widgets: list[KeyWidget]) -> QLayout:
        return build_vowel_row_containing_sets(
            build_vowel_set_layout(VOWEL_ROW_KEYS_LEFT, key_widgets),
            build_vowel_set_layout(VOWEL_ROW_KEYS_RIGHT, key_widgets),
        )


    def build_keyboard_layout(
        build_main_rows: Callable[[KeyboardWidget, list[KeyWidget]], QLayout],
        build_vowel_row: Callable[[KeyboardWidget, list[KeyWidget]], QLayout],
        key_widgets: list[KeyWidget],
    ) -> QLayout:
        """Generates a keyboard layout.
        
            :param key_widgets: A list that will be populated with the `KeyWidget`s.
        """

        # QVBoxLayout has an odd side effect when window is resized to a large enough height; the main rows layout would
        # start to grow along with the spacer item
        layout = QGridLayout()

        layout.addLayout(build_main_rows(key_widgets), 0, 0)

        layout.setRowStretch(1, 1)
        
        @watch(dpi.change)
        def resize_spacer_row():
            layout.setRowMinimumHeight(1, dpi.cm(INITIAL_ROWS_GAP_HEIGHT))
            # Allow negative stretch so the other grid rows can overlap
            QTimer.singleShot(0, lambda: layout.setRowMinimumHeight(1, -16777216))

        layout.addLayout(build_vowel_row(key_widgets), 2, 0)

        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        return layout



    right_left_width_diff = computed(lambda:
        (sum(dpi.cm(col_width.value) for col_width in col_widths_staggered_right[-N_INDEX_COLS_RIGHT:]) -
                sum(dpi.cm(col_width.value) for col_width in col_widths_staggered_left[:N_INDEX_COLS_LEFT]))
                * cos(radians(main_rows_angle.value)),
        *(width for width in col_widths_staggered_right[-N_INDEX_COLS_RIGHT:]),
        *(width for width in col_widths_staggered_left[:N_INDEX_COLS_LEFT]),
        main_rows_angle
    )


    return {
        KeyLayout.STAGGERED: partial(build_keyboard_layout, build_main_rows_layout_staggered, build_vowel_row_layout_staggered),
        KeyLayout.GRID: partial(build_keyboard_layout, build_main_rows_layout_grid, build_vowel_row_layout_grid),
    }, right_left_width_diff

#endregion


from PyQt5.QtCore import (
    Qt,
    pyqtSignal,
)
from PyQt5.QtWidgets import (
    QWidget,
    QSlider,
    QDoubleSpinBox,
)
from PyQt5.QtGui import (
    QResizeEvent,
)

from functools import wraps
from typing import Callable, Any

from ..util import on


def block_signals(fn: Callable[[Any], None]):
    @wraps(fn)
    def wrapper(self: "FloatSlider", *args):
        old_signals_blocked = self.blockSignals(True) # Prevent `valueChanged` from being fired
        fn(self, *args)
        self.blockSignals(old_signals_blocked)

    return wrapper


class FloatSlider(QSlider):
    """Slider that supports float values."""
    # Implemented by having a QSlider that can be dragged to every pixel value in its range. From there, the slider
    # values from Qt can be mapped to values in the range [0, 1]

    input = pyqtSignal(float)
    """Emitted when the user changes the value of this widget"""

    IDENTITY = lambda value: value

    def __init__(self,
        value: float=0,
        convert_in: Callable[[float], float]=IDENTITY,
        convert_out: Callable[[float], float]=IDENTITY,
        min: float=0,
        max: float=1,
        orientation: Qt.Orientation=Qt.Horizontal,
        parent: QWidget=None,
    ):
        super().__init__(orientation, parent)

        self.__convert_in = convert_in
        self.__convert_out = convert_out

        self.__min_value = self.__convert_in(min)
        self.__max_value = self.__convert_in(max)
        self.__current_value = self.__convert_in(value)

        self.__setup_ui()

    def __setup_ui(self):
        self.__update_slider_position()

        @on(self.valueChanged)
        def notify(internal_value: float):
            self.input.emit(self.__convert_out(self.__value_at(internal_value / self.__internal_max)))
        

    def resizeEvent(self, event: QResizeEvent) -> None:
        self.setMinimum(0)
        self.setMaximum(self.__internal_max)

        self.__update_slider_position()

        return super().resizeEvent(event)

    @block_signals
    def setMaximum(self, maximum: int) -> None:
        # setMaximum can emit `valueChanged` when value > maximum
        return super().setMaximum(maximum)

    @block_signals
    def setValue(self, value: int) -> None:
        return super().setValue(value)


    def __set_progress(self, progress: float):
        self.setValue(round(progress * self.__internal_max))

    def __progress_at(self, value: float):
        return (value - self.__min_value) / (self.__max_value - self.__min_value)

    def __value_at(self, progress: float):
        return progress * (self.__max_value - self.__min_value) + self.__min_value

    def __update_slider_position(self):
        self.__set_progress(self.__progress_at(self.__current_value))

    @property
    def __internal_max(self):
        """Qt's maximum value of the QSlider. (= the number of pixels it can be dragged to)"""

        if self.orientation() == Qt.Horizontal:
            return self.size().width()
        else:
            return self.size().height()

    @property
    def current_value(self):
        return self.__convert_out(self.__current_value)
    
    @current_value.setter
    def current_value(self, value: float):
        self.__current_value = self.__convert_in(value)
        self.__update_slider_position()

    @property
    def max(self):
        return self.__max_value

    @max.setter
    def max(self, value: float):
        self.__max_value = self.__convert_in(value)
        self.__update_slider_position()


class FloatEntry(QDoubleSpinBox):
    # Main purpose of this class is to define an alternative method for `setValue` that does not emit `valueChanged`

    input = pyqtSignal(float)
    """Emitted when the user changes the value of this widget"""

    def __init__(self,
        value: float=0,
        min: float=0,
        max: float=1,
        spin_step: float=0.1,
        parent: QWidget=None,
    ):
        super().__init__(parent)
        self.__setup_ui(value, min, max, spin_step)

    def __setup_ui(self, value: float, min: float, max: float, spin_step: float):
        self.setMinimum(min)
        self.setMaximum(max)
        self.setSingleStep(spin_step)
        self.current_value = value

        @on(self.valueChanged)
        def notify(value: float):
            self.input.emit(value)

    @property
    def current_value(self):
        return self.value()

    @current_value.setter
    def current_value(self, value: float):
        old_signals_blocked = self.blockSignals(True) # Prevent `valueChanged` from being fired
        self.setValue(value)
        self.blockSignals(old_signals_blocked)



from PyQt5.QtCore import (
    Qt,
    QEvent,
    pyqtSignal,
    QTimer,
    QPoint,
)
from PyQt5.QtWidgets import (
    QWidget,
    QSizePolicy,
    QGraphicsView,
)
from PyQt5.QtGui import (
    QTouchEvent,
)

from collections import Counter
from typing import cast, TYPE_CHECKING, Generator
if TYPE_CHECKING:
    from ..Main import Main
else:
    Main = object


from .KeyWidget import KeyWidget
from .RotatableKeyContainer import RotatableKeyContainer
from .build_keyboard import use_build_keyboard
from ..settings import Settings, KeyLayout
from ..util import UseDpi, Ref, watch, KEY_STYLESHEET


class KeyboardWidget(QWidget):
    end_stroke = pyqtSignal(set)  # set[str]
    current_stroke_change = pyqtSignal(set)  # set[str]
    after_touch_event = pyqtSignal()
    num_bar_pressed_change = pyqtSignal(bool)


    def __init__(self, settings: Settings, left_right_width_diff: Ref[float], parent: QWidget=None):
        super().__init__(parent)

        self.__current_stroke_keys: set[str] = set()
        self.__key_widgets: list[KeyWidget] = []

        self.__touches_to_key_widgets: dict[int, KeyWidget] = {} # keys of dict are from QTouchPoint::id
        self.__key_widget_touch_counter: Counter[KeyWidget] = Counter()

        self.settings = settings

        self.__setup_ui(left_right_width_diff)

    def __setup_ui(self, left_right_width_diff: Ref[float]):
        self.__dpi = dpi = UseDpi(self)
        build_keyboard, left_right_width_diff_src = use_build_keyboard(self.settings, self, dpi)

        self.__build_keyboard = build_keyboard
        self.setLayout(build_keyboard[self.settings.key_layout](self.__key_widgets))

        @watch(left_right_width_diff_src.change)
        def update_left_right_width_diff():
            left_right_width_diff.value = left_right_width_diff_src.value

        self.settings.key_layout_ref.change.connect(self.__rebuild_layout)


        self.setStyleSheet(KEY_STYLESHEET)

        self.setAttribute(Qt.WA_AcceptTouchEvents)
        self.setFocusPolicy(Qt.NoFocus)

        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)


    # def __handle_dpi_change(self):
    #     # self.main_rows_layout.invalidate()
    #     # self.vowel_row_layout.invalidate()
    #     # self.layout().invalidate()

    #     self.__dpi.change.emit()

    #     # self.window().setMinimumSize(self.window().sizeHint()) # Needed in order to use QWidget.resize
    #     # cast(Main, self.window()).resize_from_center(0, 0)


    def __rebuild_layout(self, value: KeyLayout):
        self.__key_widgets = []
        # Detach listeners on the old key widgets to avoid leaking memory
        # TODO removing all listeners may become overzealous in the future
        self.__dpi.change.disconnect()
        self.num_bar_pressed_change.disconnect()

        # https://stackoverflow.com/questions/10416582/replacing-layout-on-a-qwidget-with-another-layout
        QWidget().setLayout(self.layout()) # Unparent and destroy the current layout so it can be replaced
        self.setLayout(self.__build_keyboard[value](self.__key_widgets))
        

    def event(self, event: QEvent) -> bool:
        """(override)"""

        if not isinstance(event, QTouchEvent):
            return super().event(event)

        self.after_touch_event.emit()

        # Variables for detecting changes post-update
        had_num_bar = "#" in self.__current_stroke_keys

        if event.type() in (QEvent.TouchBegin, QEvent.TouchUpdate):
            old_stroke_length = len(self.__current_stroke_keys)

            self.__current_stroke_keys.update(
                key
                for key_widget in self.__find_updated_key_widgets(event.touchPoints())
                for key in key_widget.values
            )

            if len(self.__current_stroke_keys) > old_stroke_length and self.__current_stroke_keys:
                self.current_stroke_change.emit(self.__current_stroke_keys)
            if not had_num_bar and "#" in self.__current_stroke_keys:
                self.num_bar_pressed_change.emit(True)
            

        elif event.type() == QEvent.TouchEnd:
            # This also filters out empty strokes (Plover accepts them and will insert extra spaces)
            if self.__current_stroke_keys and all(touch.state() == Qt.TouchPointReleased for touch in event.touchPoints()):
                self.end_stroke.emit(set(self.__current_stroke_keys))
                self.__current_stroke_keys.clear()
                self.__touches_to_key_widgets.clear()
                self.__key_widget_touch_counter.clear()
            
            if had_num_bar:
                self.num_bar_pressed_change.emit(False)

            
        self.__update_key_widget_styles_and_state(self.__key_widget_touch_counter.keys())
        
        return True

    
    def __find_updated_key_widgets(self, touch_points: list[QTouchEvent.TouchPoint]) -> Generator[KeyWidget, None, None]:
        for touch in touch_points:
            if touch.state() == Qt.TouchPointStationary: continue

            if touch.id() in self.__touches_to_key_widgets:
                old_key_widget = self.__touches_to_key_widgets[touch.id()]
                self.__key_widget_touch_counter[old_key_widget] -= 1

                del self.__touches_to_key_widgets[touch.id()]

                if self.__key_widget_touch_counter[old_key_widget] == 0:
                    del self.__key_widget_touch_counter[old_key_widget]

            if touch.state() == Qt.TouchPointReleased: continue


            key_widget = self.__key_widget_at(touch.pos().toPoint())
            if key_widget is None: continue


            self.__touches_to_key_widgets[touch.id()] = key_widget
            self.__key_widget_touch_counter[key_widget] += 1

            
            if not key_widget.matched:
                yield key_widget


    def __key_widget_at(self, point: QPoint):
        touched_widget = self.childAt(point)
        if touched_widget is None: return

        if isinstance(touched_widget, KeyWidget):
            return touched_widget
        
        # Loop through all the key containers; since `childAt` can only return the topmost child at a point,
        # a container that obscures another container with its empty space will cause `childAt` to fail to find a
        # touched key if one is underneath. The container list is also `reversed` so that containers rendered last
        # (i.e., on top) are processed first
        for view in reversed(self.findChildren(RotatableKeyContainer)):
            view: RotatableKeyContainer = view
            key_widget: KeyWidget = view.key_widget_at_point(point)

            if key_widget is None: continue
            return key_widget

        return

        """ # For some reason, `touched_widget` is a widget that is a child of the QGraphicsView rather than the
        # QGraphicsView itself
        elif isinstance(touched_widget.parent(), RotatableKeyContainer):
            view: RotatableKeyContainer = touched_widget.parent()
            key_widget: KeyWidget = view.key_widget_at_point(point)

            if key_widget is None: return
            return key_widget """
        
        raise TypeError
        

    def __update_key_widget_styles_and_state(self, touched_key_widgets: list[KeyWidget]):
        stroke_has_ended = len(self.__key_widget_touch_counter) == 0

        for key_widget in self.__key_widgets:
            old_touched, old_matched = key_widget.touched, key_widget.matched

            if key_widget in touched_key_widgets:
                key_widget.touched = True
                key_widget.matched = True

            elif ((not stroke_has_ended and key_widget.matched) # optimization assumes keys will not be removed mid-stroke
                    or all(key in self.__current_stroke_keys for key in key_widget.values)):
                key_widget.touched = False
                key_widget.matched = True

            else:
                key_widget.touched = False
                key_widget.matched = False


            if (old_touched, old_matched) != (key_widget.touched, key_widget.matched):
                # Reload stylesheet for dynamic properties: https://stackoverflow.com/questions/1595476/are-qts-stylesheets-really-handling-dynamic-properties
                # self.style().unpolish(key_widget)
                self.style().polish(key_widget)



from PyQt5.QtCore import (
    QObject,
    pyqtSignal,
    QSettings,
)

from enum import Enum, auto
from weakref import WeakKeyDictionary
from typing import TypeVar, Generic, Any

from .util import RefAttr, on_many


T = TypeVar("T")
class _PersistentSetting(RefAttr, Generic[T]):
    """`RefAttr` whose class records which settings are on a `Settings` object so they can all be saved/loaded 
        easily.
    """

    setting_types_on_object: WeakKeyDictionary[Any, dict[str, type]] = WeakKeyDictionary()

    #region Overrides

    def __init__(self, expected_type: type[T], qsettings_type_arg: "type | None"=None):
        """
            :param qsettings_type_arg: The type object that should be passed as `type` to `QSettings::value`. If
            `NoneType` (`type(None)`), then the parameter should be omitted.
        """
        super().__init__(expected_type)
        self.__qsettings_type_arg = qsettings_type_arg if qsettings_type_arg is not None else expected_type

    def __set_name__(self, owner_class: type, attr_name: str):
        super().__set_name__(owner_class, attr_name)
        self.__attr_name = attr_name

    def __set__(self, instance: Any, value: T):
        # The owner instance is not accessible before any `__get__` and `__set__` calls, so the recording is done in
        # `__set__` (and `__set__` specifically since values are initialized in the constructor of `Settings`)
        if instance is not None:
            if instance not in _PersistentSetting.setting_types_on_object:
                _PersistentSetting.setting_types_on_object[instance] = {}
            setting_types_dict = _PersistentSetting.setting_types_on_object[instance]

            if self.__attr_name not in setting_types_dict:
                setting_types_dict[self.__attr_name] = self.__qsettings_type_arg

        return super().__set__(instance, value)

    #endregion

class KeyLayout(Enum):
    STAGGERED = auto()
    GRID = auto()


class Settings(QObject):
    key_layout = _PersistentSetting(KeyLayout, type(None))

    stroke_preview_stroke = _PersistentSetting(bool)
    stroke_preview_translation = _PersistentSetting(bool)

    key_width = _PersistentSetting(float)
    key_height = _PersistentSetting(float)
    compound_key_size = _PersistentSetting(float)

    index_stretch = _PersistentSetting(float)
    pinky_stretch = _PersistentSetting(float)

    vowel_set_offset_fac = _PersistentSetting(float)

    index_stagger_fac = _PersistentSetting(float)
    middle_stagger_fac = _PersistentSetting(float)
    ring_stagger_fac = _PersistentSetting(float)
    pinky_stagger_fac = _PersistentSetting(float)

    main_rows_angle = _PersistentSetting(float)
    vowel_rows_angle = _PersistentSetting(float)

    window_opacity = _PersistentSetting(float)


    key_layout_ref = key_layout.ref_getter()

    stroke_preview_stroke_ref = stroke_preview_stroke.ref_getter()
    stroke_preview_translation_ref = stroke_preview_translation.ref_getter()

    key_width_ref = key_width.ref_getter()
    key_height_ref = key_height.ref_getter()
    compound_key_size_ref = compound_key_size.ref_getter()

    index_stretch_ref = index_stretch.ref_getter()
    pinky_stretch_ref = pinky_stretch.ref_getter()

    vowel_set_offset_fac_ref = vowel_set_offset_fac.ref_getter()

    index_stagger_fac_ref = index_stagger_fac.ref_getter()
    middle_stagger_fac_ref = middle_stagger_fac.ref_getter() 
    ring_stagger_fac_ref = ring_stagger_fac.ref_getter() 
    pinky_stagger_fac_ref = pinky_stagger_fac.ref_getter()

    main_rows_angle_ref = main_rows_angle.ref_getter()
    vowel_rows_angle_ref = vowel_rows_angle.ref_getter()

    window_opacity_ref = window_opacity.ref_getter()


    stroke_preview_change = pyqtSignal()
    
    def __init__(self):
        super().__init__()

        self.key_layout = KeyLayout.STAGGERED

        self.stroke_preview_stroke = True
        self.stroke_preview_translation = True

        self.key_width = 1.9
        self.key_height = 2.1
        self.compound_key_size = 1.2

        self.index_stretch = -0.1
        self.pinky_stretch = 0.4

        self.vowel_set_offset_fac = 0.6

        self.index_stagger_fac = 0.5
        self.middle_stagger_fac = 0.85
        self.ring_stagger_fac = 0.5
        self.pinky_stagger_fac = 0

        self.main_rows_angle = 19
        self.vowel_rows_angle = 20.5

        self.window_opacity = 0.675

        @on_many(self.stroke_preview_stroke_ref.change, self.stroke_preview_translation_ref.change)
        def emit_stroke_preview_change():
            self.stroke_preview_change.emit()
        

    def load(self, settings: QSettings):
        for attr_name, setting_type in _PersistentSetting.setting_types_on_object[self].items():
            default_value = getattr(self, attr_name)
            if setting_type is type(None):
                setattr(self, attr_name, settings.value(attr_name, default_value))
            else:
                setattr(self, attr_name, settings.value(attr_name, default_value, type=setting_type))

    def save(self, settings: QSettings):
        for attr_name in _PersistentSetting.setting_types_on_object[self].keys():
            settings.setValue(attr_name, getattr(self, attr_name))

    @property
    def stroke_preview_full(self):
        return self.stroke_preview_stroke and self.stroke_preview_translation

    @property
    def stroke_preview_visible(self):
        return self.stroke_preview_stroke or self.stroke_preview_translation



from plover.gui_qt.engine import Engine
from plover.steno import Stroke

from plover import system
from plover.translation import Translation, Translator, _mapping_to_macro # TODO access violation

from PyQt5.QtCore import (
    Qt,
)
from PyQt5.QtWidgets import (
    QWidget,
    QLabel,
    QGridLayout,
    QVBoxLayout,
    QSpacerItem,
    QSizePolicy,
)
from PyQt5.QtGui import (
    QFont,
)


from math import cos, radians
from typing import Iterable

from ..settings import Settings, KeyLayout
from ..util import UseDpi, Ref, watch, watch_many, FONT_FAMILY

class StrokePreview(QWidget):
    def __init__(self, engine: Engine, settings: Settings, right_left_width_diff: Ref[float], parent: QWidget=None):
        super().__init__(parent)

        self.__engine = engine
        self.__settings = settings

        self.__last_translation: Translation | None = None
        self.__last_stroke_matched = True

        self.__setup_ui(right_left_width_diff)

    def __setup_ui(self, right_left_width_diff: Ref[float]):
        dpi = UseDpi(self)

        #region Labels
        self.__stroke_label = stroke_label = QLabel(self)
        self.__translation_label = translation_label = QLabel(self)

        stroke_label.setTextFormat(Qt.PlainText)
        translation_label.setTextFormat(Qt.PlainText)

        # The horizontal size policies of the labels must be set to Ignored, or otherwise they will cause the window to
        # resize to fit the text if the text is too long. However, that alone just causes the labels to resize to the
        # widest item in the QVBoxLayout `labels_layout`, which would have a width of 0 since the label widths are
        # ignored and the rest of the items are spacers. The strut is therefore needed to set the width of the layout
        # to the width of the parent column in the QGridLayout `display_alignment_layout`, allowing the labels the full
        # column width to display the text yet limiting them to that width so they do not resize the window.
        strut = QWidget(self)
        strut.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Fixed)
        # https://stackoverflow.com/questions/21739119/qt-hboxlayout-stop-mainwindow-from-resizing-to-contents
        stroke_label.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Fixed)
        translation_label.setSizePolicy(QSizePolicy.Ignored, QSizePolicy.Fixed)

        labels_layout = QVBoxLayout()
        
        labels_layout.addSpacerItem(top_spacer := QSpacerItem(0, 0))
        labels_layout.addWidget(stroke_label, 0, Qt.AlignCenter)
        labels_layout.addSpacerItem(middle_spacer := QSpacerItem(0, 0))
        labels_layout.addWidget(translation_label, 0, Qt.AlignCenter)
        labels_layout.addWidget(strut)


        @watch(dpi.change)
        def resize_label_fonts(): # Set font sizes in px rather than pt so they fit in the keyboard gaps
            stroke_label_font = QFont(FONT_FAMILY)
            translation_label_font = QFont(FONT_FAMILY)

            stroke_label_font.setPixelSize(dpi.dp(16.8 if self.__settings.stroke_preview_translation else 21))
            translation_label_font.setPixelSize(dpi.dp(21))

            stroke_label.setFont(stroke_label_font)
            middle_spacer.changeSize(0, dpi.dp(-4) if self.__settings.stroke_preview_full else 0)
            translation_label.setFont(translation_label_font)

            labels_layout.invalidate()

        labels_layout.setSpacing(0)
        #endregion


        #region Display alignment
        display_alignment_layout = QGridLayout()
        display_alignment_layout.setColumnStretch(0, 1)
        display_alignment_layout.setColumnStretch(1, 0)

        @watch_many(dpi.change, right_left_width_diff.change)
        def resize_display_alignment():
            display_alignment_layout.setColumnMinimumWidth(1, right_left_width_diff.value)
                    # dpi.cm(self.__settings.key_width) * cos(radians(self.__settings.main_rows_angle)))
            # TODO replace MAIN_ROWS_ANGLE import with setting value when this becomes a setting


        display_alignment_layout.addLayout(labels_layout, 0, 0)

        @watch_many(dpi.change, self.__settings.key_layout_ref.change)
        def reposition_labels():
            if self.__settings.key_layout == KeyLayout.GRID:
                labels_layout.setAlignment(Qt.AlignBottom)
                top_spacer.changeSize(0, 0)
            else:
                labels_layout.setAlignment(Qt.AlignTop)
                top_spacer.changeSize(0, -8)

            labels_layout.invalidate()  # Must be called for layout to move
        
        display_alignment_layout.addItem(QSpacerItem(0, 0), 0, 1)
        display_alignment_layout.setSpacing(0)
        display_alignment_layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(display_alignment_layout)
        #endregion


        #region Settings reactivity
        @watch(self.__settings.stroke_preview_change)
        def on_settings_change():
            stroke_label.setVisible(self.__settings.stroke_preview_stroke)
            translation_label.setVisible(self.__settings.stroke_preview_translation)

            resize_label_fonts()

            self.__display_translation(self.__last_translation, self.__last_stroke_matched)
            self.finish_stroke()
        #endregion


    def display_keys(self, stroke_keys: Iterable[str]):
        if not self.__settings.stroke_preview_visible:
            # The coming translation will not be computed
            # When the stroke preview is reenabled, this will cause the placeholder to be shown again instead of an old
            # translation
            self.__last_translation = None
            return

        translation, stroke_matched = _coming_translation(self.__engine, stroke_keys)
        self.__display_translation(translation, stroke_matched)


    def __display_translation(self, translation: "Translation | None", stroke_matched: bool):
        self.__last_translation = translation
        self.__last_stroke_matched = stroke_matched
        if not self.__settings.stroke_preview_visible:
            return

        if translation is None and self.__settings.stroke_preview_translation:
            self.__stroke_label.setText("")
            self.__translation_label.setText("…")
            return
        elif translation is None:
            self.__stroke_label.setText("…")
            return
            

        if self.__settings.stroke_preview_stroke:
            self.__stroke_label.setText(" / ".join(translation.rtfcre))

            self.__stroke_label.setStyleSheet("color: #41796a;")

        if self.__settings.stroke_preview_translation:
            if translation.english is not None:
                word = translation.english.replace("\n", "\\n")
            else:
                word = "[no match]" if self.__settings.stroke_preview_stroke else translation.strokes[0].rtfcre
            self.__translation_label.setText(word)
        
            self.__translation_label.setStyleSheet(f"""color: #{"ff" if stroke_matched else "5f"}41796a;""")


    def finish_stroke(self):
        self.__stroke_label.setStyleSheet("")
        self.__translation_label.setStyleSheet(f"""color: #{"ff" if self.__last_stroke_matched else "3f"}000000;""")


def _coming_translation(engine: Engine, keys: Iterable[str]) -> tuple[Translation, bool]:
    """Computes the translation that will result if the stroke defined by `keys` is sent to the engine.
    
    :returns: The translation that will result, and if a match was found
    :rtype: tuple[Translation, bool]
    """
    # TODO access violations

    translator: Translator = engine._translator
    stroke: Stroke = Stroke(keys)

    # translator._state is temporarily cleared when engine output is set to False
    if not engine.output:
        translator.set_state(engine._running_state)
    
    
    # This is mostly the body of `Translator.translate_stroke`, but without the side effects

    max_key_length = translator._dictionary.longest_key
    mapping = translator._lookup_with_prefix(max_key_length, translator.get_state().translations, [stroke])


    translation = (
        translator._find_longest_match(2, max_key_length, stroke) or
        (mapping is not None and Translation([stroke], mapping)) or
        translator._find_longest_match(1, max_key_length, stroke, system.SUFFIX_KEYS) or
        Translation([stroke], None)
    )

    stroke_matched = translation.english is not None

    if translation.english is None:
        macro = _mapping_to_macro(mapping, stroke)
        if macro is not None:
            translation = Translation([stroke], f"={macro.name}")
    

    if not engine.output:
        translator.clear_state()

    
    return translation, stroke_matched


from PyQt5.QtCore import (
    QEvent,
    QPoint,
    Qt,
    QTimer,
)
from PyQt5.QtWidgets import (
    QGraphicsView,
    QWidget,
    QLayout,
    QGraphicsProxyWidget,
    QGraphicsScene,
)
from PyQt5.QtGui import (
    QTouchEvent,
    QTransform,
    QResizeEvent,
)


from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .KeyboardWidget import KeyboardWidget
else:
    KeyboardWidget = object

from .KeyWidget import KeyWidget
from ..util import KEY_STYLESHEET

class RotatableKeyContainer(QGraphicsView):
    """
    `QGraphicsView` used to transform a group of `KeyWidget`s. Provides additional styling and access to key positions
    """

    @staticmethod
    def of_layout(layout: QLayout, align_left: bool, angle: float, parent: QWidget = None) -> "RotatableKeyContainer":
        scene = QGraphicsScene(parent)

        widget = QWidget()
        widget.setStyleSheet(KEY_STYLESHEET)
        widget.setAttribute(Qt.WA_TranslucentBackground) # Gives this container a transparent background
        widget.setLayout(layout)

        proxy = scene.addWidget(widget)

        container = RotatableKeyContainer(widget, proxy, align_left, angle, scene, parent)

        return container

    def __init__(self, widget: QWidget, proxy: QGraphicsProxyWidget, align_left: bool, angle: float, *args):
        super().__init__(*args)

        self.__widget = widget
        self.__proxy = proxy

        self.__align_left = align_left
        self.__angle = angle

        self.__setup_ui()

    def __setup_ui(self):
        # self.setAttribute(Qt.WA_AcceptTouchEvents)

        # Disable scrolling and hide scrollbars
        self.__h_scrollbar = h_scrollbar = self.horizontalScrollBar()
        self.__v_scrollbar = v_scrollbar = self.verticalScrollBar()

        h_scrollbar.setEnabled(False)
        v_scrollbar.setEnabled(False)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

        # Scroll to the bottom of the container (bottom-align the key groups)
        QTimer.singleShot(0, self.__rescroll)

        # Clear default styling
        self.setStyleSheet("background: #00000000; border: none;")

    """ def event(self, event: QEvent):
        if not isinstance(event, QTouchEvent):
            return super().event(event)

        if event.type() in (QEvent.TouchBegin, QEvent.TouchUpdate):
            screen_coords = event.touchPoints()[0].pos().toPoint()
            widget_coords = self.__proxy.deviceTransform(self.viewportTransform()).inverted()[0].map(screen_coords)
            key_widget: KeyWidget = self.__widget.childAt(widget_coords)

        elif event.type() == QEvent.TouchEnd:
            pass

        return True """

    def key_widget_at_point(self, window_coords: QPoint) -> KeyWidget:
        # Convert screen coords to container widget coords, which can then be used to retrieve the key widget

        # Undo the widget's translation
        widget_translation = QTransform()
        widget_translation.translate(self.pos().x(), self.pos().y())
        view_coords = widget_translation.inverted()[0].map(window_coords)

        widget_coords = self.__proxy.deviceTransform(self.viewportTransform()).inverted()[0].map(view_coords)

        return self.__widget.childAt(widget_coords)

    def __rescroll(self):
        self.scroll_to_bottom()
        self.scroll_to_horizontal_edge()

    def scroll_to_bottom(self):
        self.__v_scrollbar.setValue(self.__v_scrollbar.maximum())

    def scroll_to_horizontal_edge(self):
        if self.__align_left:
            self.__h_scrollbar.setValue(self.__h_scrollbar.maximum())
        else:
            self.__h_scrollbar.setValue(self.__h_scrollbar.minimum())


    def update_size(self):
        self.__update_widget_size_and_position()
        QTimer.singleShot(0, self.__update_own_size)
        # self.__update_own_size()

    def __update_widget_size_and_position(self):
        widget = self.__widget
        layout = self.__widget.layout()

        for i in range(layout.count()):
            if child_layout := layout.itemAt(i).layout():
                child_layout.invalidate()

        # Constrains the widget to the size of its layout
        widget.setMaximumSize(layout.sizeHint())

        if self.__align_left:
            self.__proxy.setPos(0, -self.__widget.geometry().height())
            self.__proxy.setTransformOriginPoint(self.__widget.rect().bottomLeft())
        else:
            self.__proxy.setPos(-self.__widget.geometry().width(), -self.__widget.geometry().height())
            self.__proxy.setTransformOriginPoint(self.__widget.rect().bottomRight())

        self.__proxy.setRotation(self.__angle)

    def __update_own_size(self):
        scene = self.scene()

        scene.setSceneRect(scene.itemsBoundingRect())
        self.setMaximumSize(scene.sceneRect().size().toSize())

        QTimer.singleShot(0, self.__rescroll)

    def resizeEvent(self, event: QResizeEvent) -> None:
        self.scroll_to_horizontal_edge()
        return super().resizeEvent(event)
    

    @property
    def angle(self):
        return self.__angle

    @angle.setter
    def angle(self, value: float):
        self.__angle = value
        self.update_size()
